﻿using DetourFunc;
using DetourFunc.Clr;
using System;
using System.Reflection;
using System.Runtime.InteropServices;
using WindowsPE;

namespace ConsoleApp1
{
    public delegate void SleepDelegate(int milliseconds);

    // Install-Package DetourFunc -Version 1.1.0
    class Program
    {
        [DllImport("kernel32.dll")]
        public static extern void Sleep(int milliseconds);

        [DllImport("kernel32", SetLastError = true, CharSet = CharSet.Ansi, BestFitMapping = false)]
        internal static extern IntPtr LoadLibrary(string lpFileName);

        [DllImport("kernel32", CharSet = CharSet.Ansi, ExactSpelling = true, SetLastError = true, BestFitMapping = false)]
        internal static extern IntPtr GetProcAddress(IntPtr hModule, string funcName);

        static SleepDelegate s_originalSleepFunc;

        static void Main(string[] _)
        {
            //{
            //    IntPtr libraryAddress = LoadLibrary("kernel32.dll");
            //    if (libraryAddress == IntPtr.Zero)
            //    {
            //        return;
            //    }

            //    IntPtr funcAddr = GetProcAddress(libraryAddress, "Sleep");
            //    Console.WriteLine($"GetProcAddress - Sleep API: {funcAddr.ToInt64():x}");
            //}

            IntPtr sleepPtr;
            {
                sleepPtr = MethodReplacer.GetExportFunctionAddress("kernel32.dll", "Sleep", out var _);
                //Console.WriteLine($"m64: {refAddress.ToInt64():x}");
                Console.WriteLine($"JumpTo KERNELBASE!Sleep: {sleepPtr.ToInt64():x}");
            }

            Console.WriteLine();

            IntPtr ptrBodyReplaceMethod;

            {
                SleepDelegate action = Replaced_TestMethod;
                MethodDesc mdReplaceMethod = MethodDesc.ReadFromMethodInfo(action.Method);
                ptrBodyReplaceMethod = mdReplaceMethod.GetNativeFunctionPointer();
            }

            Console.WriteLine($"Address to be patched: {sleepPtr.ToInt64():x}");
            Console.WriteLine($"With this address: {ptrBodyReplaceMethod.ToInt64():x}");
            Console.WriteLine();

            using (var item = new TrampolinePatch<SleepDelegate>())
            {
                if (item.JumpPatch(sleepPtr, ptrBodyReplaceMethod) == true)
                {
                    s_originalSleepFunc = item.GetOriginalFunc();
                }

                Console.WriteLine("[After trampoline]");
                Sleep(3000);
            }

            Console.WriteLine();
            Console.WriteLine("[Revert to original method]");
            Sleep(3000);

            Console.WriteLine("Press any key to exit...");
            Console.ReadLine();
        }

        public static void Replaced_TestMethod(int milliseconds)
        {
            Console.WriteLine($"Replaced_TestMethod called: milliseconds = {milliseconds}");
            s_originalSleepFunc?.Invoke(milliseconds);
        }
    }
}
